#  Drakkar-Software trading-backend
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import ccxt

import trading_backend.errors
import trading_backend.enums
import trading_backend.constants


class Exchange:
    SPOT_ID = None
    MARGIN_ID = None
    FUTURE_ID = None
    IS_SPONSORING = False

    def __init__(self, exchange):
        self._exchange = exchange

        # add backend headers
        self._exchange.connector.add_headers(self.get_headers())

    def stop(self):
        self._exchange = None

    @classmethod
    def get_name(cls):
        return 'default'

    @classmethod
    def is_sponsoring(cls) -> bool:
        return cls.IS_SPONSORING

    async def initialize(self) -> str:
        default = f"{self.get_name().capitalize()} backend initialized."
        return f"{default} {await self._ensure_broker_status()}" if self.is_sponsoring() else default

    async def _ensure_broker_status(self):
        return f"Broker rebate is enabled."

    def get_headers(self) -> dict:
        return {}

    def get_orders_parameters(self, params=None) -> dict:
        if params is None:
            params = {}
        return params

    def _allow_withdrawal_right(self) -> bool:
        return trading_backend.constants.ALLOW_WITHDRAWAL_KEYS

    async def _inner_cancel_order(self):
        # use client api to avoid any ccxt call wrapping and error handling
        await self._exchange.connector.client.cancel_order("12345", symbol="BTC/USDT")

    async def _get_api_key_rights_using_order(self) -> list[trading_backend.enums.APIKeyRights]:
        rights = [trading_backend.enums.APIKeyRights.READING]
        try:
            await self._inner_cancel_order()
        except ccxt.AuthenticationError as err:
            if "permission" in str(err).lower():
                # does not have trading permission
                pass
            else:
                # another error
                raise
        except ccxt.ExchangeError as err:
            if not self._exchange.is_api_permission_error(err):
                # has trading permission
                rights.append(trading_backend.enums.APIKeyRights.SPOT_TRADING)
                rights.append(trading_backend.enums.APIKeyRights.MARGIN_TRADING)
                rights.append(trading_backend.enums.APIKeyRights.FUTURES_TRADING)
        return rights

    async def _get_api_key_rights(self) -> list[trading_backend.enums.APIKeyRights]:
        # default implementation: fetch portfolio and don't check
        # todo implementation for each exchange as long as ccxt does not support it in unified api
        await self._exchange.connector.client.fetch_balance()
        return [
            trading_backend.enums.APIKeyRights.READING,
            trading_backend.enums.APIKeyRights.SPOT_TRADING,
            trading_backend.enums.APIKeyRights.FUTURES_TRADING,
            trading_backend.enums.APIKeyRights.MARGIN_TRADING,
        ]

    async def _ensure_api_key_rights(self):
        # raise trading_backend.errors.APIKeyPermissionsError on missing permissions
        rights = []
        try:
            rights = await self._get_api_key_rights()
        except ccxt.BaseError:
            raise
        except Exception as err:
            try:
                import octobot_commons.logging as logging
                logging.get_logger(self.__class__.__name__).exception(
                    err, True, f"Error when getting {self.__class__.__name__} api key rights: {err}"
                )
            except ImportError:
                pass
            # non ccxt error: proceed to right checks and raise
        required_right = trading_backend.enums.APIKeyRights.SPOT_TRADING
        if self._exchange.exchange_manager.is_future:
            required_right = trading_backend.enums.APIKeyRights.FUTURES_TRADING
        if self._exchange.exchange_manager.is_margin:
            required_right = trading_backend.enums.APIKeyRights.MARGIN_TRADING
        if required_right not in rights:
            raise trading_backend.errors.APIKeyPermissionsError(
                f"{required_right.value} permission is required"
            )
        if not self._allow_withdrawal_right() and trading_backend.enums.APIKeyRights.WITHDRAWALS in rights:
            raise trading_backend.errors.APIKeyPermissionsError(
                f"This api key has withdrawal rights, please revoke it."
            )

    async def is_valid_account(self, always_check_key_rights=False) -> (bool, str):
        try:
            # 1. check account
            validity, message = await self._inner_is_valid_account()
            if not always_check_key_rights and not validity:
                return validity, message
            # 2. check api key right
            await self._ensure_api_key_rights()
            return validity, message
        except trading_backend.errors.APIKeyPermissionsError:
            # forward exception
            raise
        except ccxt.InvalidNonce as err:
            raise trading_backend.errors.TimeSyncError(err)
        except ccxt.ExchangeError as err:
            raise trading_backend.errors.ExchangeAuthError(err)

    async def _inner_is_valid_account(self) -> (bool, str):
        # check account validity regarding exchange requirements, exchange specific
        return True, None

    def _get_id(self):
        if self._exchange.exchange_manager.is_future:
            return self.FUTURE_ID
        if self._exchange.exchange_manager.is_margin:
            return self.MARGIN_ID
        return self.SPOT_ID



#  Drakkar-Software trading-backend
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import ccxt.async_support
import trading_backend.exchanges as exchanges
import tests.util.create_order_tests as create_order_tests
from tests import coinex_exchange


def test_get_name(coinex_exchange):
    assert exchanges.Coinex(coinex_exchange).get_name() == ccxt.async_support.coinex().id.lower()


@pytest.mark.asyncio
async def test_get_orders_parameters(coinex_exchange):
    exchange = exchanges.Coinex(coinex_exchange)
    await create_order_tests.create_order_mocked_test_args(
        exchange,
        exchange_private_post_order_method_name="v2PrivatePostSpotOrder",
        exchange_request_referral_key="client_id",
        should_contains=True)


@pytest.mark.asyncio
async def test_sign(coinex_exchange):
    exchange = exchanges.Coinex(coinex_exchange)
    await create_order_tests.sign_test(
        exchange,
        ['v1', 'private'],
        "client_id",
        should_contains=True,
        in_body=True,
        url_path="order/limit"
    )



#  Drakkar-Software trading-backend
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import ccxt
import ccxt.static_dependencies.ecdsa.der
import binascii
import trading_backend.exchanges as exchanges
import trading_backend.enums


class Coinbase(exchanges.Exchange):
    # todo update this when coinbase broker id is available
    SPOT_ID = None
    MARGIN_ID = None
    FUTURE_ID = None
    REF_ID = None
    IS_SPONSORING = False

    @classmethod
    def get_name(cls):
        return 'coinbase'

    async def _ensure_broker_status(self):
        return f"Broker rebate is not enabled (missing broker id)."

    def _get_legacy_api_permissions(self, scopes):
        rights = []
        read_scopes = [
            "wallet:accounts:read",
            "wallet:buys:read",
            "wallet:sells:read",
            "wallet:orders:read",
            "wallet:trades:read",
            "wallet:user:read",
            "wallet:transactions:read",
        ]
        trade_scopes = [
            "wallet:buys:create",
            "wallet:sells:create",
        ]
        withdraw_scopes = [
            "wallet:withdrawals:create"
        ]
        if all(scope in scopes for scope in read_scopes):
            rights.append(trading_backend.enums.APIKeyRights.READING)
        if all(scope in scopes for scope in trade_scopes):
            rights.append(trading_backend.enums.APIKeyRights.SPOT_TRADING)
            rights.append(trading_backend.enums.APIKeyRights.MARGIN_TRADING)
            rights.append(trading_backend.enums.APIKeyRights.FUTURES_TRADING)
        if any(scope in scopes for scope in withdraw_scopes):
            rights.append(trading_backend.enums.APIKeyRights.WITHDRAWALS)
        return rights

    def _get_api_permissions(self, scopes):
        rights = []
        read_scopes = [
            "rat#view",
        ]
        trade_scopes = [
            "rat#trade",
        ]
        withdraw_scopes = [
            "rat#transfer"
        ]
        if all(scope in scopes for scope in read_scopes):
            rights.append(trading_backend.enums.APIKeyRights.READING)
        if all(scope in scopes for scope in trade_scopes):
            rights.append(trading_backend.enums.APIKeyRights.SPOT_TRADING)
            rights.append(trading_backend.enums.APIKeyRights.MARGIN_TRADING)
            rights.append(trading_backend.enums.APIKeyRights.FUTURES_TRADING)
        if any(scope in scopes for scope in withdraw_scopes):
            rights.append(trading_backend.enums.APIKeyRights.WITHDRAWALS)
        return rights

    async def _inner_cancel_order(self):
        try:
            await super()._inner_cancel_order()
        except ccxt.ArgumentsRequired as err:
            # raised on invalid api keys
            raise ccxt.AuthenticationError(err) from err

    async def _get_api_key_rights(self) -> list[trading_backend.enums.APIKeyRights]:
        # warning might become deprecated
        # https://docs.cloud.coinbase.com/sign-in-with-coinbase/docs/api-users
        try:
            return await self._get_api_key_rights_using_order()
            # DEPRECATED:
            # GET https://api.coinbase.com/v2/user/auth 410 Gone The path /v2/user/auth is deprecated and no longer exists.
            restrictions = (await self._exchange.connector.client.v2PrivateGetUserAuth())["data"]
            scopes = restrictions["scopes"]
            if rights := self._get_api_permissions(scopes):
                return rights
            # legacy api keys
            if rights := self._get_legacy_api_permissions(scopes):
                return rights
            # should not happen unless coinbase changes its api again
            self._exchange.logger.error(
                f"Can't fetch {self.get_name()} api key permissions from scopes: {scopes}. "
                f"Using default order creation check instead"
            )
            # last change: try creating an order
            return await self._get_api_key_rights_using_order()
        except ccxt.AuthenticationError:
            raise
        except (
            binascii.Error, AssertionError, IndexError,
            ccxt.ArgumentsRequired, ccxt.static_dependencies.ecdsa.der.UnexpectedDER
        ) as err:
            raise ccxt.AuthenticationError(f"Invalid key format ({err})")
        except ccxt.BaseError as err:
            self._exchange.logger.exception(
                err, True,
                f"Error when fetching {self.get_name()} api key rights: {err} ({err.__class__.__name__}). "
                f"This is not normal, endpoint might be deprecated, see"
                f"https://docs.cloud.coinbase.com/sign-in-with-coinbase/docs/api-users. "
                f"Using _get_api_key_rights_using_order() instead"
            )
            return await self._get_api_key_rights_using_order()

    async def _inner_is_valid_account(self) -> (bool, str):
        return False, await self._ensure_broker_status()



#  Drakkar-Software trading-backend
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import mock
import pytest
import ccxt.async_support
import trading_backend.exchanges as exchanges
import tests.util.create_order_tests as create_order_tests
from tests import huobi_exchange


def test_get_name(huobi_exchange):
    assert exchanges.Huobi(huobi_exchange).get_name() == ccxt.async_support.huobi().id.lower()


@pytest.mark.asyncio
async def test_get_orders_parameters(huobi_exchange):
    exchange = exchanges.Huobi(huobi_exchange)
    with mock.patch.object(exchange._exchange.connector.client,
                           "fetch_accounts", mock.AsyncMock(return_value=[{'id': 1}])):
        await create_order_tests.create_order_mocked_test_args(
            exchange,
            exchange_private_post_order_method_name="spotPrivatePostV1OrderOrdersPlace",
            exchange_request_referral_key="client-order-id",
            should_contains=True)
